use axum::{
    Router,
    routing::{get, post},
};
use instance::Instance;
use routes::{list_entries, upload_video};
use tokio::signal;
use tracing::info;

mod instance;
mod routes;
mod video;

#[tokio::main]
async fn main() -> Result<(), Box<dyn std::error::Error>> {
    tracing_subscriber::fmt::init();
    dotenvy::dotenv()?;

    let database_url = std::env::var("DATABASE_URL").unwrap_or_else(|_| "sqlite://app.db".into());
    let instance = Instance::new("almond.toml", &database_url).await?;

    info!(
        "Instance configuration:\n+ Host: {}\n+ Port: {}\n+ Password: {:?}\n+ Videos per page: {}",
        instance.config.host,
        instance.config.port,
        instance.config.password,
        instance.config.videos_per_page
    );

    let address = format!("{}:{}", instance.config.host, instance.config.port);

    let almond = Router::new()
        .route("/", get(list_entries))
        .route("/upload", post(upload_video))
        .with_state(instance);

    let listener = tokio::net::TcpListener::bind(address).await?;

    axum::serve(listener, almond.into_make_service())
        .with_graceful_shutdown(shutdown_signal())
        .await?;

    Ok(())
}

async fn shutdown_signal() {
    let ctrl_c = async {
        signal::ctrl_c()
            .await
            .expect("Failed to install Ctrl+C handler");
    };

    #[cfg(unix)]
    let terminate = async {
        signal::unix::signal(signal::unix::SignalKind::terminate())
            .expect("failed to install signal handler")
            .recv()
            .await;
    };

    #[cfg(not(unix))]
    let terminate = std::future::pending::<()>();

    tokio::select! {
        () = ctrl_c => {},
        () = terminate => {},
    }
}
